unit UnitProperties;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, VCL.Graphics, VCL.Controls,
  VCL.Forms, VCL.Dialogs, FSPlugin, VCL.StdCtrls;

type
  TfmSetup = class(TForm)
    btOk: TButton;
    btCancel: TButton;
    lbLangHeader: TLabel;
    edLang: TEdit;
    btBrowseLang: TButton;
    lbStreamsHeader: TLabel;
    cbShowStandart: TCheckBox;
    cbAlternate: TCheckBox;
    cbSecurity: TCheckBox;
    cbExtendedAttribute: TCheckBox;
    cbHardLink: TCheckBox;
    cbProperty: TCheckBox;
    cbObjectIdentifier: TCheckBox;
    cbReparsePoints: TCheckBox;
    cbSparseFile: TCheckBox;
    cbInvalid: TCheckBox;
    odOpen: TOpenDialog;
    procedure btOkClick(Sender: TObject);
    procedure btCancelClick(Sender: TObject);
    procedure btBrowseLangClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  fmSetup: TfmSetup;

  { __ Language support __________________________________________________________ }

const
  LangFileHeader = 'wfx_FileStreams language file';

var
  msg_PropCaption, msg_PropOk, msg_PropCancel, msg_PropBrowse, msg_PropStreams,
    msg_PropOpenTitle, msg_PropOpenFilter, msg_PropLangNotFound,
    msg_PropLangInvalidFile: string;

  { __ Interface _________________________________________________________________ }

var
  TC_Ver: tFsDefaultParamStruct;

const
  Flag_siInvalid = 1;
  Flag_siStandard = 2;
  Flag_siExtendedAttribute = 4;
  Flag_siSecurity = 8;
  Flag_siAlternate = 16;
  Flag_siHardLink = 32;
  Flag_siProperty = 64;
  Flag_siObjectIdentifier = 128;
  Flag_siReparsePoints = 256;
  Flag_siSparseFile = 512;

function CallProperties(MainWnd: HWnd): Boolean;

function SafePath(const APath: string; DontExpandEmptyPath: Boolean = True): string;

implementation

uses
  IniFiles, JclSysInfo, JclFileUtils;

{$R *.dfm}

function PathGetAbsolutePath(const P: string): string;
var
  CurrentDirectory: string;
begin
  if not PathIsAbsolute(P) then
  begin
    CurrentDirectory := ExtractFileDir(GetModuleName(HInstance));
    if CurrentDirectory <> '' then
      Result := PathGetRelativePath(PathAddSeparator(CurrentDirectory), P)
    else
      Result := P;
  end
  else
    Result := P;
end;

function SafePath(const APath: string; DontExpandEmptyPath: Boolean = True): string;
var
  Path: string;
begin
  Path := APath;
  if DontExpandEmptyPath and (APath = '') then
    Exit;
  ExpandEnvironmentVar(Path);
  Result := PathGetLongName(PathCanonicalize(PathGetAbsolutePath(Path)));
end;

function CallProperties(MainWnd: HWnd): Boolean;
var
  Ini: TMemIniFile;
  Streams: Integer;
  IniPath: string;
  AModalResult: TModalResult;
begin
  Application.Handle := MainWnd;
  fmSetup := TfmSetup.Create(nil);
  try
    if msg_PropCaption <> '' then
      fmSetup.Caption := msg_PropCaption;
    if msg_PropOk <> '' then
      fmSetup.btOk.Caption := msg_PropOk;
    if msg_PropCancel <> '' then
      fmSetup.btCancel.Caption := msg_PropCancel;
    //if msg_PropBrowse <> '' then
      //fmSetup.btBrowseLang.Hint := msg_PropBrowse;
    if msg_PropStreams <> '' then
      fmSetup.lbStreamsHeader.Caption := msg_PropStreams;
    if msg_PropOpenTitle <> '' then
      fmSetup.odOpen.Title := msg_PropOpenTitle;
    if msg_PropOpenFilter <> '' then
      fmSetup.odOpen.Filter := msg_PropOpenFilter;
    Streams := -1;
    try
      if TC_Ver.DefaultIniName <> '' then
        IniPath := TC_Ver.DefaultIniName
      else
        IniPath := ExtractFilePath(GetModuleName(HInstance)) + 'TC_wfx_NTFSFileStreams.ini';

      Ini := TMemIniFile.Create(IniPath);
      try
        fmSetup.edLang.Text := SafePath(Ini.ReadString(
          'wfx_NTFSFileStreams', 'Language', ''));
        Streams := Ini.ReadInteger('wfx_NTFSFileStreams',
          'StreamTypes', Streams);
      finally
        FreeAndNil(Ini);
      end;
    except
      on E: Exception do;
    end;
    fmSetup.cbInvalid.Checked := Streams and Flag_siInvalid <> 0;
    fmSetup.cbShowStandart.Checked := Streams and Flag_siStandard <> 0;
    fmSetup.cbExtendedAttribute.Checked := Streams and
      Flag_siExtendedAttribute <> 0;
    fmSetup.cbSecurity.Checked := Streams and Flag_siSecurity <> 0;
    fmSetup.cbAlternate.Checked := Streams and Flag_siAlternate <> 0;
    fmSetup.cbHardLink.Checked := Streams and Flag_siHardLink <> 0;
    fmSetup.cbProperty.Checked := Streams and Flag_siProperty <> 0;
    fmSetup.cbObjectIdentifier.Checked := Streams and
      Flag_siObjectIdentifier <> 0;
    fmSetup.cbReparsePoints.Checked := Streams and Flag_siReparsePoints <> 0;
    fmSetup.cbSparseFile.Checked := Streams and Flag_siSparseFile <> 0;

    AModalResult := fmSetup.ShowModal;
    Result := AModalResult = mrOk;
  finally
    FreeAndNil(fmSetup);
  end;
end;

procedure TfmSetup.btOkClick(Sender: TObject);
var
  Ini: TMemIniFile;
  Streams: Integer;
  S, IniPath: string;
  SL: TStringList;
begin
  if edLang.Text <> '' then
  begin
    if not FileExists(edLang.Text) then
    begin
      Application.MessageBox(PChar(msg_PropLangNotFound), PChar(Caption),
        mb_Ok or mb_IconInformation);
      Exit;
    end;

    try
      {
      FS := TFileStream.Create(edLang.Text, fmOpenRead or fmShareDenyWrite);
      try
        SetLength(S, Length(LangFileHeader));
        if FS.read(S[1], Length(S)) < Length(S) then
          raise Exception.Create('');
        if S <> LangFileHeader then
          raise Exception.Create('');
      finally
        FreeAndNil(FS);
      end;
      }
      SL := TStringList.Create;
      try
        SL.LoadFromFile(edLang.Text);
        if SL.Count > 0 then
          S := SL[0]
        else
          S := '';
        if S <> LangFileHeader then
          raise Exception.Create('');
      finally
        FreeAndNil(SL);
      end;
    except
      on E: Exception do
      begin
        Application.MessageBox(PChar(msg_PropLangInvalidFile), PChar(Caption),
          mb_Ok or mb_IconInformation);
        Exit;
      end;
    end;
  end;

  Streams := 0;
  if cbInvalid.Checked then
    Streams := Streams or Flag_siInvalid;
  if cbShowStandart.Checked then
    Streams := Streams or Flag_siStandard;
  if cbExtendedAttribute.Checked then
    Streams := Streams or Flag_siExtendedAttribute;
  if cbSecurity.Checked then
    Streams := Streams or Flag_siSecurity;
  if cbAlternate.Checked then
    Streams := Streams or Flag_siAlternate;
  if cbHardLink.Checked then
    Streams := Streams or Flag_siHardLink;
  if cbProperty.Checked then
    Streams := Streams or Flag_siProperty;
  if cbObjectIdentifier.Checked then
    Streams := Streams or Flag_siObjectIdentifier;
  if cbReparsePoints.Checked then
    Streams := Streams or Flag_siReparsePoints;
  if cbSparseFile.Checked then
    Streams := Streams or Flag_siSparseFile;
  try
    if TC_Ver.DefaultIniName <> '' then
      IniPath := TC_Ver.DefaultIniName
    else
      IniPath := ExtractFilePath(GetModuleName(HInstance)) + 'TC_wfx_NTFSFileStreams.ini';

    Ini := TMemIniFile.Create(IniPath);
    try
      Ini.WriteString('wfx_NTFSFileStreams', 'Language',
        ExtractRelativePath(ExtractFilePath(GetModuleName(HInstance)), edLang.Text));
      Ini.WriteInteger('wfx_NTFSFileStreams', 'StreamTypes', Streams);
      Ini.UpdateFile;
    finally
      FreeAndNil(Ini);
    end;
  except
    on E: Exception do;
  end;
  ModalResult := mrOk;
  //Close;
end;

procedure TfmSetup.btCancelClick(Sender: TObject);
begin
  ModalResult := mrCancel;
  Close;
end;

procedure TfmSetup.btBrowseLangClick(Sender: TObject);
begin
  odOpen.InitialDir := ExtractFileDir(GetModuleName(HInstance));
  if edLang.Text = '' then
    odOpen.FileName := ChangeFileExt
      (GetModuleName(HInstance), '.lng')
  else
    odOpen.FileName := edLang.Text;
  if odOpen.Execute then
    edLang.Text := odOpen.FileName;
end;

end.
